function curv = Curvatures(xc, smear, ends, curvSkip);

% Computes the row vector of curvatures for the shape xc
% xc is a row vector in the complex representation

% The curvatures are invariant to size, alignment and rotation 
% of the shape, hence the curvatures
% are insensitve to the procrustes alignment and needs to be recomputed 
% only for the single shape changed.

j = sqrt(-1);

xp = xc([2:end, 1]);
xm = xc([end, 1:(end-1)]);

x2der = xp + xm - 2*xc;
xperp = -j * (xp - xm);
xperpNorm2 = real(xperp.*conj(xperp)); 
                    % real needed to cope with lack of precision

curvFac = length(xc)  / pi;
curv = curvFac * real(x2der .* conj(xperp)) ./ xperpNorm2;

if(ends~=1) curv = curv(2:(end-1)); end;

if smear>0
    t = ceil(2*smear);
    indi = 2 * t + 1;
    filx = -t:1:t;
    fil = exp(-0.5 * filx.^2 / smear^2);
    fil = fil / sum(fil);
    
    if(ends~=1) 
        curvExt = [curv(1) * ones(1, t), curv, curv(end) * ones(1, t)];
    else
        curvExt = [curv(end-t+1 : end), curv, curv(1:t)];
    end
    curvFil = conv(curvExt, fil);
    curv = curvFil(2*t+1 : end-2*t);  
end

if(ends~=1) curv = curv(curvSkip : end-curvSkip+1); end;

    

    








