function [A, MDL, stepsizes, mu, V, lambdas, AHis, MDLHis] = ...
    Optimise(C, Cn, Ccum, ends, nPasses, P, A, ps, npsOpt, ...
    stepsizes, cutoff, curvSmear, curvFac, curvSkip)

% Optimisation of shape reparametrisation wrt description length
% For each stepsize, try + and - 
%    record the number of success changes and reduce step size 
%    when less than successPct

mohammed = 1;
maxSteps = 5;
successPct = 5;
successPct2 = 20;
counter = 0;

extras = ends; 
nps = length(ps);

s = size(C, 1);

MDLHis = zeros(1, nPasses*npsOpt*maxSteps*s);
AHis = zeros(s, nPasses*maxSteps, nps);
for ss=1:s
    for node=1:nps
        AHis(ss,:,node) = A(ss,ps(node));
    end
end

it = 0;

if ends==1, N = P; else,  N = P + 1; end
X = zeros(s, N);

if curvFac>0
    NK = N;
    if(ends~=1) NK = N - 2 * CurvSkip; end;
    K = zeros(s, NK);
end

for ss=1:s
    X(ss,:) = FormShape(C(ss,1:Cn(ss)), Ccum(ss,1:Cn(ss)), A(ss,:), P, ends);
    X(ss,:) = X(ss,:) - mean(X(ss,:));
    X(ss,:) = X(ss,:) / norm(X(ss,:));
    if curvFac>0 
        K(ss,:) = Curvatures(X(ss,:), curvSmear, ends, curvSkip);
    end
end
Xa = Procrustes(X);
lambdas = PCAShapes(Xa);
MDL = DescriptionLength(lambdas, cutoff);
if curvFac>0 MDLK = curvFac * CurvCost(K); else MDLK = 0; end
MDL = MDL + MDLK;
if curvFac>0 disp('Pass Step Cost .... Cost-terms ....     log stepsize');
else disp('Pass Step Cost     log stepsize'); end
disp(sprintf('start      %4.4f', MDL));

for pass = 1:nPasses
    for node = 1:npsOpt
        p = ps(node);
        
        for nstep=1:maxSteps
            successes = 0;
            for ss = 1:s
                if(ss==1 & ~mohammed) continue, end;
                for stepsign = [-1, 1]
                    step = stepsizes(node) * stepsign;
                    if(ss>1)
                        atest = A(ss,:);
                        atest(p) = A(ss,p) + step;
                        if ends==2 & p>=P
                            if atest(p)<0, continue, end
                        end
                        xsave = X(ss,:);
                        if curvFac>0 ksave = K(ss,:); end
                        X(ss,:) = FormShape(C(ss,1:Cn(ss)), Ccum(ss,1:Cn(ss)), ...
                               atest, P, ends);
                        X(ss,:) = X(ss,:) - mean(X(ss,:));
                        X(ss,:) = X(ss,:) / norm(X(ss,:));
                        if curvFac>0 
                            K(ss,:) = Curvatures(X(ss,:), curvSmear, ends, curvSkip);
                        end
                   else    % mohammed case
                        Atest = A;
                        Atest(2:s,p) = A(2:s,p) + step/2;
                        if ends==2 & p>=P
                            if any(Atest(:,p)<0), continue, end
                        end
                        Xsave = X;
                        if curvFac>0 Ksave = K; end
                        for sss=2:s
                            X(sss,:) = FormShape(C(sss,1:Cn(sss)), ...
                                Ccum(sss,1:Cn(sss)), Atest(sss,:), P, ends);
                            X(sss,:) = X(sss,:) - mean(X(sss,:));
                            X(sss,:) = X(sss,:) / norm(X(sss,:));
                            if curvFac>0 
                                K(sss,:) = Curvatures(X(sss,:), curvSmear, ends, curvSkip);
                            end
                        end
                    end

                    Xa = Procrustes(X);
                    lambdas = PCAShapes(Xa);
                    DL = DescriptionLength(lambdas, cutoff);
                    if curvFac>0 DLK = curvFac * CurvCost(K); else DLK = 0; end
                    DL = DL + DLK;
                    counter = counter + 1;
                    if DL<MDL 
                        if(ss>1)
                            A(ss,p) = atest(p);
                            successes = successes + 1;
                        else % mohammed case
                            A(:,p) = Atest(:,p);
                            successes = successes + s/2;
                        end
                        
                        MDL = DL;
                        MDLK = DLK;
                        break;
                    else
                        if(ss>1)
                            X(ss,:) = xsave;
                            if curvFac>0 K(ss,:) = ksave; end
                        else 
                            X = Xsave;
                            if curvFac>0 K = Ksave; end
                        end
                    end
                end % of sign loop
                it = it + 1;
                MDLHis(it) = MDL;
                AHis(ss, maxSteps*(pass-1) + nstep, node) = A(ss,p);
            end % of example loop
            
            if(successes/s < successPct/100) ...
                    stepsizes(node) = stepsizes(node)/2;end
            if(successes/s > successPct2/100) ...
                    stepsizes(node) = stepsizes(node)*2;end       
        end % of step loop
        if curvFac>0 disp(sprintf('%3d %3d    %4.4f %4.4f %4.4f  %2.6f', ...
                pass, node, MDL, MDL-MDLK, MDLK, ...
                mean( log(stepsizes) )/log(10) ));
        else disp(sprintf('%3d %3d    %4.4f  %2.6f', ...
                pass, node, MDL, ...
                mean( log(stepsizes) )/log(10) ));
        end
       
    end  % of node loop
end % of pass loop

Xa = Procrustes(X);
[lambdas, V, mu] = PCAShapes(Xa);
disp(sprintf('Number of trials: %d', counter));
