function [A, MDL, stepsizes, aTargetNow, mu, V, lambdas, AHis, MDLHis] = ...
    OptimiseCon(C, Cn, Ccum, ends, nPasses, P, A, ps, npsOpt, ...
    stepsizes, aTargetNow, cutoff, mode, aTarget, tolerance, ...
    curvSmear, curvFac, CurvSkip)

% Optimisation of shape reparametrisation wrt description length
% For each stepsize, try + and - perturbations of node parameters,
%    record the number of success changes and reduce step size 
%    when there are less than successPct
% Version with NodeCost for full control of run away
%
% This version avoids the original "mode 1" and supports mode 2-4:
%
% mode 2: Simply use NodeCost with target = the given aTarget
% mode 3: Use moving target so that example 1 stays at aTarget
% mode 4: Use moving target so that average stays at aTarget
%    or in other words
% mode 2: aTarget close to average
% mode 3: aTarget = example 1
% mode 4: aTarget = average
%--------------------------------------------------------------  

maxSteps = 5;
successPct = 5;
successPct2 = 20;
counter = 0;

extras = ends; 
nps = length(ps);

s = size(C, 1);

MDLHis = zeros(1, nPasses*npsOpt*maxSteps*s);
AHis = zeros(s, nPasses*maxSteps, nps);
for ss=1:s
    for node=1:nps
        AHis(ss,:,node) = A(ss,ps(node));
    end
end

it = 0;

if ends==1, N = P; else,  N = P + 1; end
X = zeros(s, N);

if curvFac>0 
    NK = N;
    if(ends~=1) NK = N - 2 * CurvSkip; end;
    K = zeros(s, NK);
end

for ss=1:s
    X(ss,:) = FormShape(C(ss,1:Cn(ss)), Ccum(ss,1:Cn(ss)), A(ss,:), P, ends);
    X(ss,:) = X(ss,:) - mean(X(ss,:));
    X(ss,:) = X(ss,:) / norm(X(ss,:));
    if curvFac>0 
        K(ss,:) = Curvatures(X(ss,:), curvSmear, ends, CurvSkip); 
    end
end
Xa = Procrustes(X);
lambdas = PCAShapes(Xa);
if curvFac>0 MDLK = curvFac * CurvCost(K); else MDLK = 0; end

Mpenalty = NodePenalty(A, ps, aTargetNow, tolerance);
MDL = DescriptionLength(lambdas, cutoff) + Mpenalty;
MDL = MDL + MDLK;
disp('Pass Step Cost .... Cost-terms ....     log stepsize');
disp(sprintf('start      %4.4f %4.4f', MDL, Mpenalty));

for pass = 1:nPasses
    
    % Adjust moving targets    
    if mode>2 
        if mode==3
            aObs = A(1,ps);
        else
            aObs = mean(A(:,ps));
        end
        aShift = aTarget - aObs;
        aTargetNow = aTargetNow + aShift;
        
        % ..and adjust the parameters and recompute shapes    
        
        A(:,ps) = A(:,ps) + ones(s,1) * aShift;
        for ss=1:s
            X(ss,:) = FormShape(C(ss,1:Cn(ss)), Ccum(ss,1:Cn(ss)), ...
                A(ss,:), P, ends);
            X(ss,:) = X(ss,:) - mean(X(ss,:));
            X(ss,:) = X(ss,:) / norm(X(ss,:));
            if curvFac>0 
                K(ss,:) = Curvatures(X(ss,:), curvSmear, ends, CurvSkip); 
            end
        end
        Xa = Procrustes(X);
        lambdas = PCAShapes(Xa);
        
        if curvFac>0 MDLK = curvFac * CurvCost(K);else MDLK=0; end
        Mpenalty = NodePenalty(A, ps, aTargetNow, tolerance);
        MDL = DescriptionLength(lambdas, cutoff) + Mpenalty + MDLK;
    end
    
    % Do the optimaisation for this pass (4 layer loop)
    
    for node = 1:npsOpt
        p = ps(node);
        
        for nstep=1:maxSteps
            successes = 0;
            for ss = 1:s
                for stepsign = [-1, 1]
                    step = stepsizes(node) * stepsign;
                    atest = A(ss,:);
                    atest(p) = A(ss,p) + step;
                    if ends==2 & p>=P
                        if atest(p)<0, continue, end
                    end
                    xsave = X(ss,:);
                    if curvFac>0 ksave = K(ss,:); end
                    X(ss,:) = FormShape(C(ss,1:Cn(ss)), Ccum(ss,1:Cn(ss)), ...
                        atest, P, ends);
                    X(ss,:) = X(ss,:) - mean(X(ss,:));
                    X(ss,:) = X(ss,:) / norm(X(ss,:));
                    if curvFac>0 
                        K(ss,:) = Curvatures(X(ss,:), curvSmear, ends, CurvSkip);
                    end
                    Xa = Procrustes(X);
                    lambdas = PCAShapes(Xa);
                    
                    if curvFac>0 DLK = curvFac*CurvCost(K);else DLK=0; end
                    penalty = NodePenalty(A, ps, aTargetNow, tolerance);
                    DL = DescriptionLength(lambdas, cutoff) + penalty + DLK;
                    counter = counter + 1;
                    if DL<MDL 
                        A(ss,p) = atest(p);
                        
                        MDL = DL;
                        MDLK = DLK;
                        Mpenalty = penalty;
                        successes = successes + 1;
                        break;
                    else
                        X(ss,:) = xsave;
                        if curvFac>0 K(ss,:) = ksave;end
                    end
                end
                it = it + 1;
                MDLHis(it) = MDL;
                AHis(ss, maxSteps*(pass-1) + nstep, node) = A(ss,p);
            end
            
            if(successes/s < successPct/100) ...
                    stepsizes(node) = stepsizes(node)/2;end
            if(successes/s > successPct2/100) ...
                    stepsizes(node) = stepsizes(node)*2;end                  
        end
     
        if curvFac>0 disp(sprintf('%3d %3d  %4.4f  %4.4f %4.4f %4.4f  %2.6f', ...
                pass, node, MDL, MDL-MDLK-Mpenalty, MDLK, Mpenalty, ...
                mean( log(stepsizes) )/log(10) ));          
        else disp(sprintf('%3d %3d  %4.4f  %4.4f %4.4f  %2.6f', ...
                pass, node, MDL, MDL-Mpenalty, Mpenalty, ...
                mean( log(stepsizes) )/log(10) ));      
        end
    end
end

Xa = Procrustes(X);
[lambdas, V, mu] = PCAShapes(Xa);
disp(sprintf('Number of trials: %d', counter));
