function Xa = Procrustes(X);

% Aligns 2D shapes delivered as 
%   comples rows of X and puts them in Xa
% Shapes are assumed to be centred and normalised -
% The algorithm is based on chapter 3 of
%    Statistical Shape Analysis
%    I. L. Dryden, Kanti V. Mardia
%    Wiley 1998
%
% S is n by n, n=64 is typical, s=24 is typical
% so often is may be better to diagonalise an s by s matrix
%  formed by the transvserse trick. 
% 

[s,n] = size(X);

if n<s  
%     disp('standard procrustes');
         
    S1 = X' * X; 
    S = S1 + S1';  
           % this is needed to avoid numerical problem in eig
    
% a slower alternative
%     S = zeros(n);
%     for ss=1:s
%         S = S + X(ss,:)' * X(ss,:);  
%     end 
    
%     S = conj(S); % to agree with Mardia
%     [V,D] = eig(S);
%     muR = V(:,end).';  % transpose with no conj
    
    [V,D] = eig(S);
    muR = V(:,end)';
else
%     disp('transverse procrustes');
    
    St1 = X * X'; 
    St = St1 + St1';  
           % this is needed to avoid numerical problem in eig
%     St1c = St1'
%     figure(33);
%     plot(St1(:) - St1c(:));
   
% a slower alternative
%     St = zeros(s);
%     for nn=1:n
%         St = St + X(:,nn) * X(:,nn)';
%     end
    
    [V,D] = eig(St);
    muR  = V(:,end)' * X; %  (X' * V(:,end))'
    muR = muR / norm(muR);
    
end

% rotate muR along X(1,:)
%  theta = angle(X(1,:) * muR');  
%           % The one which moves suffers conj i.e. '
% mu = muR * exp(j*theta);
% plot(X(1,:),'r-+'); axis equal;
% hold on;
% plot(muR,'b-+'); axis equal;
% plot(mu,'g-+'); axis equal;
% hold off;

mu = muR * (X(1,:) * muR') ;

% for ss=1:s
%     Xa(ss,:) = (mu * X(ss,:)') * X(ss,:);
% end   - now replaced by the following faster code

rotvec = mu * X';
Xa = X .* repmat(rotvec.', 1, n);





 